// file      : xsd/cxx/parser/parser-template.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-template.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      //
      //
      struct StartElement : Traversal::Element,
                            protected virtual Context
      {
        StartElement (Context& c, String const& u_base)
            : Context (c), u_base_ (u_base)
        {
        }

        virtual Void
        traverse (Type& e)
        {
	  if (skip (e)) return;

          String const& name (ename (e));
          String inst (L"_xsd_" + name + L"__");

          if (e.qualified ())
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && " <<
              "ns == " << L << "\"" << e.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && ns.empty ())"
               << "{";
          }

          os << "this->" << u_base_ << "::context_.top ().parser_ = " << inst << ";"
             << endl
             << "if (this->" << inst << ")"
             << "{"
             << "this->" << inst << "->pre ();"
             << "}"
             << "return true;"
             << "}";
        }

      private:
        String u_base_;
      };


      //
      //
      struct EndElement : Traversal::Element,
                          protected virtual Context
      {
        EndElement (Context& c, String const& name, String const& u_base)
            : Context (c), name_ (name), u_base_ (u_base)
        {
        }

        virtual Void
        traverse (Type& e)
        {
	  if (skip (e)) return;

          String const& name (ename (e));
          String inst (L"_xsd_" + name + L"__");
          String base_type (L"_xsd_" + name_ + L"::_xsd_" + name +
                            L"< _xsd_" + name + L"_ >");

          if (e.qualified ())
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && " <<
              "ns == " << L << "\"" << e.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && ns.empty ())"
               << "{";
          }

          //os << u_base_ << "::parser_ = 0;"
          //   << endl

          os << "if (this->" << inst << ")"
             << "{"
             << "::xsd::cxx::parser::glue< _xsd_" << name << "_ >::call (" <<
            endl
             << "(" << base_type << "*) (this)," << endl
             << "&" << base_type << "::" << name << "," << endl
             << "this->" << inst << "," << endl
             << "&" << parser_template << "< _xsd_" << name << "_, " <<
            char_type << " >::post);"
            //<< "this->" << name << " (this->" << inst << "->post ());"
             << "}"
             << "return true;"
             << "}";
        }

      private:
        String name_;
        String u_base_;
      };


      //
      //
      struct Attribute : Traversal::Attribute,
                         protected virtual Context
      {
        Attribute (Context& c, String const& name)
            : Context (c), name_ (name)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));
          String inst (L"_xsd_" + name + L"__");

          if (a.qualified ())
          {
            os << "if (n == " << L << "\"" << a.name () << "\" && " <<
              "ns == " << L << "\"" << a.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << a.name () << "\" && ns.empty ())"
               << "{";
          }

          String base_type (L"_xsd_" + name_ + L"::_xsd_" + name +
                            L"< _xsd_" + name + L"_ >");

          os << "if (this->" << inst << ")"
             << "{"
             << "this->" << inst << "->pre ();"
             << "this->" << inst << "->_characters (v);"

             << "::xsd::cxx::parser::glue< _xsd_" << name << "_ >::call (" <<
            endl
             << "(" << base_type << "*) (this)," << endl
             << "&" << base_type << "::" << name << "," << endl
             << "this->" << inst << "," << endl
             << "&" << parser_template << "< _xsd_" << name << "_, " <<
            char_type << " >::post);"
            //<< "this->" << name << " (this->" << inst << "->post ());"
             << "}"
             << "return true;"
             << "}";
        }

      private:
        String name_;
      };


      //
      //
      struct MemberStruct : Traversal::Member,
                            protected virtual Context
      {
        MemberStruct (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String targ (unclash (name, "x"));

          os << "template < typename " << targ << " >" << endl
             << inl
             << "_xsd_" << name << "< " << targ << " >::" << endl
             << "~_xsd_" << name << " () {}";

           os << "template < typename " << targ << " >" << endl
             << inl
             << "void _xsd_" << name << "< " << targ << " >::" << endl
             << name << " (" << targ << " const&) {}";

          os << "template <>" << endl
             << "struct _xsd_" << name << "<void>"
             << "{"
             << "virtual ~_xsd_" << name << " () {}"
             << "virtual void " << name << " () {}"
             << "};";
        }
      };

      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              member_struct_ (c)
        {
          names_member_struct_ >> member_struct_;
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean he (has<Traversal::Element> (c));
          Boolean ha (has<Traversal::Attribute> (c));

          if (!(he || ha))
            return;

          String const& name (ename (c));
          String type (L"_xsd_" + name + L"_");

          // Ultimate base type.
          String u_base (L"::xsd::cxx::parser::event_consumer_impl< " +
                         char_type + L" >");

          os << "// " << name << endl
             << "//" << endl
             << endl;

          os << "namespace _xsd_" << name
             << "{";

          //
          //
          names (c, names_member_struct_);


          //
          //
          if (he)
          {
            // _start_element_impl
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.traverse (c);
            }

            os << " >" << endl
               << "bool " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.traverse (c);
            }

            os << "::" << endl
               << "_start_element_impl (" << string_type << " const& ns," <<
              endl
               << string_type << " const& n)"
               << "{"
               << "if (";

            {
              BaseType base_type (*this, name, false);
              Traversal::Inherits inherits_type;
              inherits_type >> base_type;

              inherits<Complex> (
                c, inherits_type, 0, 0, &Complex::default_base);
            }

            os << "::_start_element_impl (ns, n))"
               << "{"
               << "return true;"
               << "}";

            {
              StartElement start_element_ (*this, u_base);
              Traversal::Names start_element_names_;

              start_element_names_ >> start_element_;

              names (c, start_element_names_);
            }

            os << "return false;"
               << "}";


            // _end_element_impl
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.traverse (c);
            }

            os << " >" << endl
               << "bool " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.traverse (c);
            }


            os << "::" << endl
               << "_end_element_impl (" << string_type << " const& ns," <<
              endl
               << string_type << " const& n)"
               << "{"
               << "if (";

            {
              BaseType base_type (*this, name, false);
              Traversal::Inherits inherits_type;
              inherits_type >> base_type;

              inherits<Complex> (
                c, inherits_type, 0, 0, &Complex::default_base);
            }

            os << "::_end_element_impl (ns, n))"
               << "{"
               << "return true;"
               << "}";

            {
              EndElement end_element_ (*this, name, u_base);
              Traversal::Names end_element_names_;

              end_element_names_ >> end_element_;

              names (c, end_element_names_);
            }

            os << "return false;"
               << "}";
          }


          if (ha)
          {
            // _attribute_impl
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.traverse (c);
            }

            os << " >" << endl
               << "bool " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.traverse (c);
            }

            os << "::" << endl
               << "_attribute_impl (" << string_type << " const& ns," << endl
               << string_type << " const& n," << endl
               << string_type << " const& v)"
               << "{"
               << "if (";

            {
              BaseType base_type (*this, name, false);
              Traversal::Inherits inherits_type;
              inherits_type >> base_type;

              inherits<Complex> (
                c, inherits_type, 0, 0, &Complex::default_base);
            }

            os << "::_attribute_impl (ns, n, v))"
               << "{"
               << "return true;"
               << "}";

            {
              Attribute attribute (*this, name);
              Traversal::Names names_attribute;
              names_attribute >> attribute;

              names (c, names_attribute);
            }

            os << "return false;"
               << "}";
          }

          os << "}" // namespace
             << endl;
        }

        virtual Void
        default_base (Type& c)
        {
          os << "::xsd::cxx::parser::event_consumer_impl< " <<
            char_type << " >";
        }

      private:
        //
        //
        MemberStruct member_struct_;
        Traversal::Names names_member_struct_;
      };
    }

    Void
    generate_parser_template (Context& ctx, SemanticGraph::Schema& tu)
    {
      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      Complex complex (ctx);

      names >> complex;

      schema.dispatch (tu);
    }
  }
}

